<?xml version="1.0"?>

<!-- Transforms XQuery test files into Java programs with JUnit tests. -->

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="2.0"
  >

  <xsl:output method="text" encoding="UTF-8" omit-xml-declaration="yes"/>
  <xsl:param name="package"/>
  <xsl:param name="javaMaxString" select="32000"/>

  <xsl:template match="xquery-tests">
    <xsl:choose>
      <xsl:when test="run">
        <xsl:for-each select="run">
          <xsl:choose>
            <xsl:when test="@name">
              <xsl:variable name="class-name" select="concat(substring-before(../@name, 'Test'), @name, 'Test')"/>
              <xsl:result-document href="{ concat($class-name, '.java') }" omit-xml-declaration="yes">
                <xsl:call-template name="generate-test-class">
                  <xsl:with-param name="class-name" select="$class-name"/>
                </xsl:call-template>
              </xsl:result-document>
            </xsl:when>
            <xsl:otherwise>
              <xsl:call-template name="generate-test-class">
                <xsl:with-param name="class-name" select="string(../@name)"/>
              </xsl:call-template>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </xsl:when>
      <xsl:otherwise>
        <xsl:call-template name="generate-test-class">
          <xsl:with-param name="class-name" select="string(@name)"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="generate-test-class">
    <xsl:param name="class-name"/>
    /* Generated file, do not edit. */
    package <xsl:value-of select="$package"/>;

    import java.io.ByteArrayOutputStream;
    import java.io.PrintStream;
    import java.util.Iterator;
    
    import org.w3c.dom.Node;
    import org.w3c.dom.ls.LSSerializer;
    
    import com.xhive.dom.interfaces.XhiveLibraryChildIf;
    import com.xhive.dom.interfaces.XhiveDocumentIf;
    import com.xhive.dom.interfaces.XhiveLibraryIf;
    import com.xhive.index.interfaces.XhiveIndexIf;
    import com.xhive.dom.DomNode;
    import com.xhive.error.XhiveException;
    import com.xhive.dom.DomException;
    import com.xhive.query.interfaces.XhiveXQueryParam;
    import com.xhive.query.interfaces.XhiveXQueryValueIf;
    import com.xhive.query.interfaces.XhiveXQueryQueryIf;

    import com.xhive.testframework.XhiveXQueryTestCase;

    public class <xsl:value-of select="$class-name"/> extends XhiveXQueryTestCase {
    private XhiveLibraryChildIf document;
    private XhiveDocumentIf creation;

    public <xsl:value-of select="$class-name"/>(String name) {
    super(name);
    }

    public void setUp() throws Exception {
    super.setUp();
    XhiveLibraryChildIf doc;
    XhiveLibraryIf lib = getDatabase().getRoot();
    <xsl:apply-templates select="resourceLocation"/>
    <xsl:apply-templates mode="load" select="library | document | setup_versioning"/>
    doc = lib;
    <xsl:call-template name="indexes"/>
    commitAndBegin();
    doc = getDatabase().getRoot();
    <xsl:apply-templates mode="get" select=".//library[.//@primary='true']"/>
    <xsl:apply-templates mode="get" select=".//document[@primary='true']"/>
    document = doc;
    <xsl:if test=".//document[@creation]">
      doc = getDatabase().getRoot();
      <xsl:apply-templates mode="get" select=".//library[.//document/@creation='true']"/>
      <xsl:apply-templates mode="get" select=".//document[@creation='true']"/>
      creation = (XhiveDocumentIf)doc;
    </xsl:if>
    }
    <xsl:apply-templates mode="setup" select=".//library"/>

    <xsl:apply-templates select="//test">
      <xsl:with-param name="class-name" select="$class-name"/>
    </xsl:apply-templates>
    }
  </xsl:template>

  <xsl:template mode="setup" match="library">
    public void setupLibrary_<xsl:value-of select="@name"/>(XhiveLibraryIf parent) throws Exception {
    XhiveLibraryChildIf doc;
    XhiveLibraryIf lib = loadLibrary(parent, &quot;<xsl:value-of select="@name"/>&quot;
    <xsl:if test="@lockwithparent">, <xsl:value-of select="@lockwithparent"/></xsl:if>);
    <xsl:apply-templates mode="load" select="library | document"/>
    doc = lib;
    <xsl:call-template name="indexes"/>
    }
  </xsl:template>

  <xsl:template mode="load" match="library">
    setupLibrary_<xsl:value-of select="@name"/>(lib);
  </xsl:template>

  <xsl:template mode="load" match="setup_versioning">
    /* When called, setup a default versioned system */
    setupVersioning();
  </xsl:template>

  <xsl:template mode="load" match="document">
    doc = loadDocument(lib, &quot;<xsl:value-of select="file"/>&quot;,
    <xsl:value-of select="@namespaces"/>,
    <xsl:value-of select="@validation"/>);
    <xsl:apply-templates select="metadata"/>
    
    <!-- This is a workaround for a bug in Saxon 8B
         If we call the template directly, then $class-name in 
         "generate-test-class" gets corrupted somehow.
      -->
    <!--xsl:call-template name="indexes"/-->
    <xsl:apply-templates select="valueindex"/>
    <xsl:apply-templates select="fulltextindex"/>
    <xsl:apply-templates select="elementnameindex"/>
    <xsl:apply-templates select="selectedelementnameindex"/>
    <xsl:apply-templates select="metadatavalueindex"/>
    <xsl:apply-templates select="metadatafulltextindex"/>
    <xsl:apply-templates select="pathvalueindex"/>
  </xsl:template>

  <xsl:template match="metadata">
    doc.getMetadata().put(&quot;<xsl:value-of select="@key"/>&quot;,
    <xsl:choose>
      <xsl:when test="@value">
        &quot;<xsl:value-of select="@value"/>&quot;
      </xsl:when>
      <xsl:otherwise>
        null
      </xsl:otherwise>
    </xsl:choose>
    );
  </xsl:template>

  <xsl:template name="indexes">
    <xsl:apply-templates select="valueindex"/>
    <xsl:apply-templates select="fulltextindex"/>
    <xsl:apply-templates select="elementnameindex"/>
    <xsl:apply-templates select="selectedelementnameindex"/>
    <xsl:apply-templates select="metadatavalueindex"/>
    <xsl:apply-templates select="metadatafulltextindex"/>
    <xsl:apply-templates select="pathvalueindex"/>
  </xsl:template>

  <xsl:template match="valueindex">
    doc.getIndexList().addValueIndex(
    <xsl:call-template name="put-index-names"/>
    <xsl:call-template name="put-index-options"/>);
  </xsl:template>

  <xsl:template match="fulltextindex">
    doc.getIndexList().addFullTextIndex(
    <xsl:call-template name="put-index-names"/>
    <xsl:call-template name="put-index-analyzer"/>
    <xsl:call-template name="put-index-options"/>);
  </xsl:template>

  <xsl:template match="metadatavalueindex">
    doc.getIndexList().addMetadataValueIndex(
    &quot;<xsl:value-of select="indexname"/>&quot;,
    <xsl:choose>
      <xsl:when test="keyname">
        &quot;<xsl:value-of select="keyname"/>&quot;
      </xsl:when>
      <xsl:otherwise>
        null
      </xsl:otherwise>
    </xsl:choose>
    <xsl:call-template name="put-index-options"/>);
  </xsl:template>

  <xsl:template match="metadatafulltextindex">
    doc.getIndexList().addMetadataFullTextIndex(
    &quot;<xsl:value-of select="indexname"/>&quot;,
    <xsl:choose>
      <xsl:when test="keyname">
        &quot;<xsl:value-of select="keyname"/>&quot;
      </xsl:when>
      <xsl:otherwise>
        null
      </xsl:otherwise>
    </xsl:choose>
    <xsl:call-template name="put-index-analyzer"/>
    <xsl:call-template name="put-index-options"/>);
  </xsl:template>

  <xsl:template match="pathvalueindex">
    doc.getIndexList().addPathValueIndex(
    &quot;<xsl:value-of select="indexname"/>&quot;,
    &quot;<xsl:value-of select="path"/>&quot;
    <xsl:call-template name="put-index-options"/>);
  </xsl:template>

  <xsl:template name="put-index-analyzer">
    <xsl:choose>
      <xsl:when test="analyzer">
        ,&quot;<xsl:value-of select="."/>&quot;
      </xsl:when>
      <xsl:otherwise>
        ,null
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="put-index-names">
    &quot;<xsl:value-of select="indexname"/>&quot;,
    <xsl:choose>
      <xsl:when test="elementname">
        <xsl:choose>
          <xsl:when test="elementname/uri">
            &quot;<xsl:value-of select="elementname/uri"/>&quot;,
          </xsl:when>
          <xsl:otherwise>
            null,
          </xsl:otherwise>
        </xsl:choose>
        &quot;<xsl:value-of select="elementname/localname"/>&quot;,
      </xsl:when>
      <xsl:otherwise>
        null, null,
      </xsl:otherwise>
    </xsl:choose>
    <xsl:choose>
      <xsl:when test="attributename">
        <xsl:choose>
          <xsl:when test="attributename/uri">
            &quot;<xsl:value-of select="attributename/uri"/>&quot;,
          </xsl:when>
          <xsl:otherwise>
            null,
          </xsl:otherwise>
        </xsl:choose>
        &quot;<xsl:value-of select="attributename/localname"/>&quot;
      </xsl:when>
      <xsl:otherwise>
        null, null
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="put-index-options">
    , XhiveIndexIf.KEY_SORTED
    <xsl:if test="option">
      <xsl:for-each select="option">
        | XhiveIndexIf.<xsl:value-of select="."/>
      </xsl:for-each>
    </xsl:if>
  </xsl:template>

  <xsl:template match="elementnameindex">
    doc.getIndexList().addElementNameIndex(
    &quot;<xsl:value-of select="indexname"/>&quot;
    );
  </xsl:template>

  <xsl:template match="selectedelementnameindex">
    doc.getIndexList().addElementNameIndex(
    &quot;<xsl:value-of select="indexname"/>&quot;,
    new String[] {
    <xsl:apply-templates mode="selectfornameindex" select="elementname"/>
    }
    );
  </xsl:template>

  <xsl:template mode="selectfornameindex" match="elementname">
    &quot;<xsl:if test="uri"><xsl:value-of select="uri"/><xsl:text> </xsl:text></xsl:if><xsl:value-of
      select="localname"/>&quot;<xsl:if test="following-sibling::elementname">,</xsl:if>
  </xsl:template>

  <xsl:template mode="get" match="library">
    doc = (XhiveLibraryIf)((XhiveLibraryIf)doc).get(&quot;<xsl:value-of select="@name"/>&quot;);
  </xsl:template>

  <xsl:template mode="get" match="document">
    doc = (XhiveDocumentIf)((XhiveLibraryIf)doc).get(&quot;<xsl:value-of select="file"/>&quot;);
  </xsl:template>

  <xsl:template match="test">
    <xsl:param name="class-name"/>
    /**
    * &lt;author>Automatically generated&lt;/author>
    * &lt;description><xsl:choose>
      <xsl:when test="description">
        <xsl:value-of select="description"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="query"/>
      </xsl:otherwise>
    </xsl:choose>&lt;/description>
    */
    public void
    <xsl:choose><xsl:when test="@name">test<xsl:call-template name="replace">
      <xsl:with-param name="string" select="@name"/>
      <xsl:with-param name="pattern" select="'-'"/>
      <xsl:with-param name="replacement" select="'_'"/>
      </xsl:call-template></xsl:when><xsl:otherwise>testXQuery<xsl:number/></xsl:otherwise>
    </xsl:choose>
    () throws Throwable {
    boolean failure = <xsl:value-of select="@fails = 'true' or contains(@fails, concat($class-name,','))"/>;
    <xsl:if test="stdout">
      PrintStream oldOut = System.out;
      ByteArrayOutputStream newArrayOut = new ByteArrayOutputStream();
      PrintStream newOut = new PrintStream(newArrayOut);
      oldOut.flush();
      System.setOut(newOut);
    </xsl:if>
    <xsl:if test="stderr">
      PrintStream oldOut = System.err;
      ByteArrayOutputStream newArrayOut = new ByteArrayOutputStream();
      PrintStream newOut = new PrintStream(newArrayOut);
      oldOut.flush();
      System.setErr(newOut);
    </xsl:if>
    try {
    <xsl:choose>
      <xsl:when test="exception">
        try {
        <xsl:if test="modification">
          XhiveXQueryQueryIf mod_query = document.createXQuery(
          &quot;<xsl:call-template name="java-escape">
            <xsl:with-param name="string" select="modification"/>
          </xsl:call-template>&quot;);
          mod_query.setCreationDocument(creation);
          <xsl:for-each select="param">
            mod_query.setVariable(&quot;<xsl:value-of select="@name"/>&quot;,
            <xsl:value-of select="."/>);
          </xsl:for-each>
          mod_query.execute();
        </xsl:if>
        XhiveXQueryQueryIf query = document.createXQuery(
        &quot;<xsl:call-template name="java-escape">
          <xsl:with-param name="string" select="query"/>
        </xsl:call-template>&quot;);
        query.setCreationDocument(creation);
        <xsl:for-each select="param">
          query.setVariable(&quot;<xsl:value-of select="@name"/>&quot;,
          <xsl:value-of select="."/>);
        </xsl:for-each>
        Iterator result = query.execute();
        while (result.hasNext()) result.next();
        fail(&quot;Expected exception with code <xsl:value-of select="exception/@code"/>&quot;);
        } 
        <xsl:if test="exception"> catch (XhiveException e) {
          <xsl:if test="exception/@code">
          assertEquals(&quot;X-Hive Exception code should be <xsl:value-of select="exception/@code"/>&quot;,
            XhiveException.<xsl:value-of select="exception[1]/@code"/>,
            e.getErrorCode());</xsl:if>
        } catch (DomException de) {
          <xsl:if test="exception[@type='DOM']/@code">assertEquals(&quot;DOM Exception type should be <xsl:value-of select="exception/@code"/>&quot;,
            DomException.<xsl:value-of select="exception[1]/@code"/>, de.type);</xsl:if>
        }</xsl:if>
      </xsl:when>
      <xsl:otherwise>
        <xsl:if test="modification">
          XhiveXQueryQueryIf mod_query = document.createXQuery(
          &quot;<xsl:call-template name="java-escape">
            <xsl:with-param name="string" select="modification"/>
          </xsl:call-template>&quot;);
          mod_query.setCreationDocument(creation);
          <xsl:for-each select="param">
            mod_query.setVariable(&quot;<xsl:value-of select="@name"/>&quot;,
            <xsl:value-of select="."/>);
          </xsl:for-each>
          mod_query.execute();
        </xsl:if>
        XhiveXQueryQueryIf query = document.createXQuery(
        &quot;<xsl:call-template name="java-escape">
          <xsl:with-param name="string" select="query"/>
        </xsl:call-template>&quot;);
        query.setCreationDocument(creation);
        <xsl:for-each select="param">
          query.setVariable(&quot;<xsl:value-of select="@name"/>&quot;,
          <xsl:value-of select="."/>);
        </xsl:for-each>
        Iterator result = query.execute();
        XhiveXQueryValueIf value;
        <xsl:apply-templates select="value"/>
        <xsl:apply-templates select="serializedResult" />
        if (result.hasNext()) {
        fail("Result sequence should be exhausted, found: " +
        result.next().toString());
        }
      </xsl:otherwise>
    </xsl:choose>
    <xsl:if test="stdout | stderr">
      newOut.flush();
      <xsl:choose>
        <xsl:when test="( stdout | stderr )/@mode='exact'">
          assertEquals("Text to stdout/stderr should match",
          insertNewlines(&quot;<xsl:call-template name="java-escape">
            <xsl:with-param name="string" select="( stdout | stderr )"/>
          </xsl:call-template>&quot;),
          newArrayOut.toString());
        </xsl:when>
        <xsl:when test="( stdout | stderr )/@mode='contains'">
          assertContains("Text to stdout/stderr should contain expected string",
          newArrayOut.toString(),
          insertNewlines(&quot;<xsl:call-template name="java-escape">
            <xsl:with-param name="string" select="( stdout | stderr )"/>
          </xsl:call-template>&quot;));
        </xsl:when>
      </xsl:choose>
    </xsl:if>
    if (failure) {
    failure = false;
    fail("This test was expected to fail");
    }
    } catch (Throwable t) {
    if (!failure) throw t;
    <xsl:if test="stdout">
      } finally {
      System.setOut(oldOut);
    </xsl:if>
    <xsl:if test="stderr">
      } finally {
      System.setErr(oldOut);
    </xsl:if>
    }
    }
  </xsl:template>

  <xsl:template match="value">
    assertTrue("There should be another value (<xsl:number/>)", result.hasNext());
    value = (XhiveXQueryValueIf)result.next();
    <xsl:if test="@type">assertEquals(&quot;Type <xsl:number/> of value &quot; +
    value.toString() +
    &quot; should be <xsl:value-of select="@type"/>&quot;,
    DomNode.<xsl:value-of select="@type"/>,
      value.getNodeType());</xsl:if>
    assertEquals("Value <xsl:number/> should match",
    &quot;<xsl:call-template name="java-escape">
      <xsl:with-param name="string" select="text()"/>
    </xsl:call-template>&quot;, value.toString());
  </xsl:template>

  <xsl:template match="serializedResult">
    LSSerializer seri = getDatabase().getRoot().createLSSerializer();
    StringBuilder resultBuf = new StringBuilder();
    while (result.hasNext()) {
      value = (XhiveXQueryValueIf) result.next();
      if (value.isNode()) {
        resultBuf.append(seri.writeToString((Node) value));
      } else {
        resultBuf.append(value.toString());
      <xsl:choose>
        <!--xsl:when test="@source = 'xml'">if (result.hasNext()) resultBuf.append("\n");</xsl:when-->
        <xsl:when test="@compare = 'Text'">if (result.hasNext()) resultBuf.append(" ");</xsl:when>
      </xsl:choose>
      }
    }
    
    <xsl:choose>
      <xsl:when test="string-length(text()) > $javaMaxString">
        <xsl:call-template name="stringPartEcho">
          <xsl:with-param name="number" select="floor(string-length(text()) div $javaMaxString + 0.5)"/>
          <xsl:with-param name="string" select="text()"/>
        </xsl:call-template>
        String expRes = 
        <xsl:call-template name="collectParts">
          <xsl:with-param name="number" select="0"/>
          <xsl:with-param name="max" select="floor(string-length(text()) div $javaMaxString + 0.5)"/>
        </xsl:call-template>;
      </xsl:when>
      <xsl:otherwise>
        String expRes = &quot;<xsl:call-template name="java-escape">
          <xsl:with-param name="string" select="text()"/>
        </xsl:call-template>&quot;;
      </xsl:otherwise>
    </xsl:choose>;
    <xsl:choose>
      <xsl:when test="@compare ='Text'">
        assertEquals(&quot;Serialized textresult should match&quot;, 
          expRes.trim(), resultBuf.toString().trim());
      </xsl:when>
      <xsl:when test="@compare = 'XML'">
        assertXMLEquals(&quot;XML result should match&quot;,
          expRes, resultBuf.toString());
      </xsl:when>
      <xsl:when test="@compare = 'Fragment'">
        assertXMLFragmentEquals(&quot;XML fragment result should match&quot;,
          expRes, resultBuf.toString());
      </xsl:when>
      <xsl:when test="@compare = 'Ignore'">
        // no-op, result ignored
      </xsl:when>
      <xsl:when test="@compare = 'Inspect'">
        // no-op, result has to be human-inspected
      </xsl:when>
    </xsl:choose>
  </xsl:template>
  
  <xsl:template name="stringPartEcho">
    <xsl:param name="number"/>
    <xsl:param name="string"/>
    <xsl:if test="$number > -1">
      String subExpRes<xsl:value-of select="$number"/> =  &quot;<xsl:call-template name="java-escape">
        <xsl:with-param name="string" select="substring($string, $number * $javaMaxString, $javaMaxString)"></xsl:with-param>
      </xsl:call-template>&quot;;
      <xsl:call-template name="stringPartEcho">
        <xsl:with-param name="number" select="$number -1 "/>
        <xsl:with-param name="string" select="substring($string, 0, $number * $javaMaxString)"/>
      </xsl:call-template>
    </xsl:if>
  </xsl:template>
  
  <xsl:template name="collectParts">
    <xsl:param name="number"/>
    <xsl:param name="max"/>
    <xsl:if test="$number &lt; $max">subExpRes<xsl:value-of select="$number"/> + <xsl:call-template name="collectParts">
        <xsl:with-param name="max" select="$max"/>
        <xsl:with-param name="number" select="$number + 1"/>
      </xsl:call-template></xsl:if>
    <xsl:if test="$number = $max">subExpRes<xsl:value-of select="$number"/></xsl:if>
  </xsl:template>

  <xsl:template name="java-escape">
    <xsl:param name="string"/>
    <xsl:if test="string-length($string) > 0">
    <xsl:choose>
      <xsl:when test="string-length($string) = 1">
        <xsl:call-template name="java-escape1">
          <xsl:with-param name="char" select="$string"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:call-template name="java-escape">
          <xsl:with-param name="string" select="substring($string, 1, string-length($string) div 2)"/>
        </xsl:call-template>
        <xsl:call-template name="java-escape">
          <xsl:with-param name="string" select="substring($string, (string-length($string) div 2) +1)"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
    </xsl:if>
  </xsl:template>

  <xsl:template name="java-escape1">
    <xsl:param name="char"/>
    <xsl:choose>
      <xsl:when test="$char = '\'">
        <xsl:text>\\</xsl:text>
      </xsl:when>
      <xsl:when test="$char = '&quot;'">
        <xsl:text>\&quot;</xsl:text>
      </xsl:when>
      <xsl:when test="$char = '&#xA;'">
        <xsl:text>\n&quot; + &#xA;      &quot;</xsl:text>
      </xsl:when>
      <xsl:when test="$char = '&#xD;'">
        <xsl:text>\r</xsl:text>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$char"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
  
  <xsl:template match="resourceLocation">
    setResourceLocation("<xsl:value-of select="."/>");
  </xsl:template>

  
  <xsl:template name="replace">
    <xsl:param name="string" select="''"/>
    <xsl:param name="pattern" select="''"/>
    <xsl:param name="replacement" select="''"/>
    <xsl:choose>
      <xsl:when test="$pattern != '' and $string != '' and contains($string, $pattern)">
        <xsl:value-of select="substring-before($string, $pattern)"/>
        <!--
Use "xsl:copy-of" instead of "xsl:value-of" so that users
may substitute nodes as well as strings for $replacement.
-->
        <xsl:copy-of select="$replacement"/>
        <xsl:call-template name="replace">
          <xsl:with-param name="string" select="substring-after($string, $pattern)"/>
          <xsl:with-param name="pattern" select="$pattern"/>
          <xsl:with-param name="replacement" select="$replacement"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$string"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

</xsl:stylesheet>
